/*
  J+ (c) lallous <lallousx86(at)yahoo.com>

  History
  ---------
  03/23/2004 06:16 PM
    * Initial version: 0.1

  04/18/2004 12:25 PM
    * Added new function address display (after setting newfirstsegment) to the "." output
    * Now initially newfirstsegment = firstsegment

  09/23/2004 09:44 PM
    * added to "?" to display also "here" in the newfirstsegment terms
    * added "len" command
    * version 0.2

  09/26/2004 04:27 PM
    * made "wb" and "rb"

  ToDo
  ------
  * Add to default jumping facility to detect name and jump to them as if "G"
  * '^' jump to first x-ref of a given function
  * add a command that will print all (n first) xrefs and their names in the msg window
  * allow stacking of function names -> keep on adding names till he presses "w+" so that all functions get written to disk?
  * enable "?" to work even if not in a function, and let it display some other infos! like .seg+xyz

  Knowning issues/Limitations
  ------------------------------
  * not checking if some commands are issued within segments (or disabled addresses)

*/

#include <idc.idc>
#include <common.idc>

// define indexes of global variables here
#define g_newfirstsegment 0

/*
  Usage
  --------
  "=address"       : Sets the new image base.
  "."              : Goes to top of function.
  "_"              : Goes to end of function.
  "?"              : Computes size of function and different information about the function
  "address"        : Goes to address in VA form. A newfirstsegment value must be set first.
                     VA does not have to be true in the database, J+ will compute new VA.
  "+bytes"         : Goes to current segmentVA+bytes. Useful if you have addresses of the form:
                     .text+1234 or .data+2121
  "w"              : Generates functionname.asm file into database's directory.
  ">"              : Goes to next function
  "<"              : Goes to previous function
  ".."             : Goes to segment's start
  "__"             : Goes to segment's end
  "i"              : Displays internal information
  "rb filename"    : Reads the filename into the current selection. If no selection then into whole function
  "wb filename"    : Writes to filename the current selection. If no selection then the whole function is writte
  "undo"           : Restore's original function's bytes
  "len"            : Displays the lengths of a given selection or function
*/

static jplus()
{
  auto functionstart, functionend, functionname, 
       x, i, s, newfirstsegment, cmd, temp, 
       insideafunction, haveselection,
       selstart, selend;
  
  newfirstsegment = GetGlobalVarLong(g_newfirstsegment);

  insideafunction = FunctionStart(ScreenEA()) != BADADDR;

  selstart = SelStart();
  selend   = SelEnd();

  // get function's start
  functionstart = FunctionStart(ScreenEA());

  // get function's end
  functionend = FindFuncEnd(functionstart);

  haveselection = (selstart != BADADDR) && (selend != BADADDR);

  // ask for a command
  s = AskStr("", "Address");

  // get first character
  cmd = substr(s, 0, 1);

  // set new image base?
  if (cmd == "=")
  {
    s = substr(s, 1, -1);

    newfirstsegment = xtol(s);
    // set global variable
    SetGlobalVarLong(g_newfirstsegment, newfirstsegment);
    Message("New first segment %08X specified\n", newfirstsegment);
    return;
  }
  // displays internal information
  else if (cmd == "i")
  {
    Message("> New first segment: %X\n", newfirstsegment);
    return;
  }
  // the LEN command
  else if (s == "len")
  {
    // inside a function, but without any selection
    // then we write whole function
    if (insideafunction && !haveselection)
    {
      Message("here! %X-%X", functionstart, functionend);
      selstart = functionstart;
      selend   = functionend;
    }
    // not inside a function and w/o a selection
    else if (!haveselection)
    {
      Message("No selection!\n");
      return;
    }
    x = selend - selstart;
    Message("> Selection (%X-%X) length: %d bytes, %d words, %d dwords %d lines\n", selend, selstart, x, x / 2, x / 4, HowManyLines(selstart, selend));
  }
  else if (substr(s, 0, 2) == "wb")
  {
    // no selection for "wb" ?
    if (!haveselection && insideafunction)
    {
      selstart = functionstart;
      selend   = functionend;
    }
    else if (!haveselection)
    {
      Message("No selection!\n");
      return;
    }
    // extract file name
    temp = substr(s, 3, -1);
    if (temp == "")
    {
      Message("No file name specified!");
      return 0;
    }
    x = fopen(temp, "wb");
    savefile(x, 0, selstart, selend - selstart);
    fclose(x);
    x = selend - selstart;
    Message("> Wrote to file from %X-%X of %d/%X bytes", selstart, selend, x, x);
  }
  else if (substr(s, 0, 2) == "rb")
  {
    // no selection for "rb" ?
    if (!haveselection && insideafunction)
    {
      selstart = functionstart;
      selend   = functionend;
    }
    else if (!haveselection)
    {
      Message("No selection!\n");
      return;
    }
    // extract file name
    temp = substr(s, 3, -1);
    if (temp == "")
    {
      Message("No file name specified!");
      return;
    }
    x = fopen(temp, "rb"); 
    if (x == 0)
    {
      Message("> Failed to load %s!\n", temp);
      return;
    }
    loadfile(x, 0, selstart, selend - selstart);
    fclose(x);
    x = selend - selstart;
    Message("> Read from file to %X-%X %d/%X bytes", selstart, selend, x, x);
  }
  // relative jump in current segment
  else if (cmd == "+")
  {
    // parse address from full command
    x = xtol(substr(s, 1, -1));
    if (x <= 0)
    {
      Message("Invalid relative address to jump to!");
      return;
    }
    x = x + SegStart(ScreenEA());
    Message("Relative jump to %08X\n", x);
    Jump(x);
  }
  // function related commands?
  else if ((s == ".") || (s == "_") || (s == "?") || (s == "w") || (s == "undo"))
  {
    // get function's name
    functionname = GetFunctionName(functionstart);

    if (functionstart == BADADDR)
    {
      Message("Not in a function!\n");
      return;
    }
    // go to start of function?
    if (s == ".")
      Jump(functionstart);
    else if (s == "_")
    {
      x = PrevHead(functionend, functionstart);
      Jump(x);
    }
    // compute function size
    else if (s == "?")
    {
      x = functionend - functionstart;
      if (newfirstsegment)
        i = functionstart - SegStart(here) + newfirstsegment;
      else
        i = functionstart;

      Message("%s(@%0X) BYTES: (%d, 0x%X) LINES: (%d) CURRENT VA: (%X)\n", 
            functionname, i, x, x, 
            HowManyLines(functionstart, functionend),
            (here - SegStart(here) + newfirstsegment)
           );
    }
    // write function to ASM file
    else if (s == "w")
    {
      x = GetFilePath(GetIdbPath()) + functionname+".asm";

      if (WriteTxt(x, functionstart, functionend) == -1)
        Message("Failed to write to %s\n", x);
      else
        Message("Successfully generated %s\n", x);
    }
    // undoes a previously applied patch
    else if (s == "undo")
    {
      cmd = 0;
      for (i = functionstart; i <= functionend; i++)
      {
        x = GetOriginalByte(i);
        if (x != Byte(i))
        {
          PatchByte(i, x);
          cmd++;
        }
      }
      if (cmd)
      {
        s = form("Restored %d byte(s) in %s()\nDo you want to re-analyze this function?", cmd, functionname);
        if (AskYN(0, s) == 1)
          AnalyseArea(functionstart, functionend);
      }
      else
        Message("%s() was never patched!\n", functionname);
    }
    return;
  }

  // go to top of previous function
  else if (s == "<")
    Jump(PrevFunction(here));

  // go to next function
  else if (s == ">")
    Jump(NextFunction(here));

  // go to top of segment
  else if (s == "..")
    Jump(SegStart(here));

  else if (s == "__")
    Jump(SegEnd(here));
  // plain hex address entered
  else
  {
    x = xtol(s);
    if (newfirstsegment == 0)
    {
      Message("New first segment not set! Please set it first!\n");
      return;
    }
    // compute difference
    x = x - newfirstsegment;
    if (x < 0)
    {
      Message("Invalid jump address!\n");
      return;
    }
    x = x + FirstSeg();
    Jump(x);
    Message("Jumping to VA: %08X\n", x);
    return;
  }
}

static main()
{
  InitGlobalVars();

  SetGlobalVarLong(g_newfirstsegment, FirstSeg());

  AddHotkey("J", "jplus");

  Message("J+ v0.2 - loaded, press 'J' to start\n");
}